from argparse import ArgumentParser, RawDescriptionHelpFormatter

from mediakit.info import name, version, description
from mediakit.streams.colors import colored, Colors


def create_argument_parser():
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        description=(
            colored(
                f'{name.lower()} v{version}\n',
                style=Colors.style.BRIGHT
            )
            + description
        )
    )

    return parser


def parse_download_arguments():
    parser = create_argument_parser()

    parser.add_argument(
        'video_url',
        help='URL of the YouTube video to download'
    )
    parser.add_argument(
        'output_path',
        nargs='?',
        default='./',
        help='Destination folder to where to save the downloads'
    )
    parser.add_argument(
        '-f',
        '--formats',
        nargs='*',
        default=[],
        help='Formats to download, separated by spaces (e.g. 1080p 720p 360p)'
    )

    arguments = parser.parse_args()

    return arguments



from os import path

from mediakit.utils.files import increment_filename_if_exists
from mediakit.utils.commands import run_command_in_background
from mediakit.constants import FFMPEG_BINARY


VIDEO_FORMATS = { 'mp4' }


class ConversionOptions:
    NO_AUDIO = '-an'


def merge_video_and_audio(
    video_path,
    audio_path,
    output_file_path,
    output_format='mp4'
):
    final_output_file_path = increment_filename_if_exists(output_file_path)

    command = (
        f'{FFMPEG_BINARY} -i "{video_path}" -i "{audio_path}" '
        f'-vcodec copy -f {output_format} '
        f'"{final_output_file_path}"'
    )

    run_command_in_background(command)


def convert_media(file_path, output_file_path, output_format, options=[]):
    final_output_file_path = increment_filename_if_exists(output_file_path)

    command = (
        f'{FFMPEG_BINARY} -i "{file_path}" '
        + ('-vcodec copy ' if output_format in VIDEO_FORMATS else '')
        + f'-f {output_format} '
        + f'"{final_output_file_path}" '
        + ' '.join(options)
    )

    run_command_in_background(command)



import colorama


class Colors():
    fore = colorama.Fore
    back = colorama.Back
    style = colorama.Style


class ColoredText():
    def __init__(self, text, fore, back, style):
        self.text = text
        self.fore = fore
        self.back = back
        self.style = style

    def __str__(self):
        return (
            f'{self.fore}{self.back}{self.style}'
            f'{self.text}'
            f'{Colors.style.RESET_ALL}'
        )


def colored(
    *values,
    fore=Colors.fore.RESET,
    back=Colors.back.RESET,
    style=Colors.style.NORMAL,
    sep=' '
):
    joined_text = sep.join(map(
        lambda value: str(value),
        values
    ))

    return str(ColoredText(joined_text, fore, back, style))



from os import devnull
from shutil import which
import subprocess


def run_command_in_background(command):
    with open(devnull, 'w') as DEVNULL:
        subprocess.run(
            command,
            shell=True,
            stdout=DEVNULL,
            stderr=subprocess.STDOUT,
            check=True
        )


def is_command_available(command):
    return which(command) is not None



name = 'MediaKit'
version = '0.4.2'
description = 'Download YouTube videos fast, directly from the command line'
author = 'Diego Aquino'
author_email = 'diegocruzdeaquino@gmail.com'
url = 'https://github.com/diego-aquino/mediakit'
download_url = 'https://github.com/diego-aquino/mediakit/archive/v0.4.2.tar.gz'
license = 'MIT'
copyright = 'Copyright 2020 Diego Aquino'

temp_filename = 'mediakit_temp'



from .utils.ffmpeg import get_ffmpeg_binary


FFMPEG_BINARY = get_ffmpeg_binary()

DOWNLOAD_FORMATS = { 'videoaudio', 'audio', 'videoonly' }

VIDEO_DEFINITIONS = {
    'max': { 'next': '2160p' },
    '2160p': { 'next': '1440p' },
    '1440p': { 'next': '1080p' },
    '1080p': { 'next': '720p' },
    '720p': { 'next': '480p' },
    '480p': { 'next': '360p' },
    '360p': { 'next': '240p' },
    '240p': { 'next': '144p' },
    '144p': { 'next': None }
}

VIDEO_DEFINITIONS_ALIASES = {
    '4k': '2160p'
}

AUDIO_DEFINITIONS = {
    'max': { 'next': '160kbps' },
    '160kbps': { 'next': '128kbps' },
    '128kbps': { 'next': '70kbps' },
    '70kbps': { 'next': '50kbps' },
    '50kbps': { 'next': None }
}

AVAILABLE_DEFINITIONS = set()
for definition_group in [
    VIDEO_DEFINITIONS,
    VIDEO_DEFINITIONS_ALIASES,
    AUDIO_DEFINITIONS
]:
    for definition in definition_group:
        AVAILABLE_DEFINITIONS.add(definition)



import os

from imageio.plugins import ffmpeg

from mediakit.utils.commands import is_command_available


def get_ffmpeg_binary():
    ffmpeg_binary = (
        _try_global_command()
        or _try_existing_binary()
        or _try_imageio_ffmpeg()
    )

    return ffmpeg_binary


def _try_global_command():
    if is_command_available('ffmpeg'):
        return 'ffmpeg'
    if is_command_available('ffmpeg.exe'):
        return 'ffmpeg.exe'

    return None


def _try_existing_binary():
    ffmpeg_binary = os.getenv('FFMPEG_BINARY')

    if ffmpeg_binary and is_command_available(ffmpeg_binary):
        return ffmpeg_binary

    return None


def _try_imageio_ffmpeg():
    imageio_ffmpeg_binary = ffmpeg.get_exe()

    if is_command_available(imageio_ffmpeg_binary):
        return imageio_ffmpeg_binary

    return None



from os import path

from mediakit.info import temp_filename
from mediakit.media.convert import (
    merge_video_and_audio,
    convert_media,
    ConversionOptions
)
from mediakit.utils.files import get_safe_filename, remove_file, move_file
from mediakit.constants import VIDEO_DEFINITIONS_ALIASES


class DownloadStatusCodes():
    READY = 'READY'
    DOWNLOADING = 'DOWNLOADING'
    CONVERTING = 'CONVERTING'
    DONE = 'DONE'


class MediaResource:
    def __init__(
        self,
        source,
        output_type,
        output_path=None,
        definition='max',
        filename='',
        filename_suffix=''
    ):
        self.source = source
        self.output_type = output_type
        self.output_path = output_path

        final_definition = (
            VIDEO_DEFINITIONS_ALIASES.get(definition, definition)
            if output_type.startswith('video')
            else definition
        )
        is_alias_definition = definition != final_definition

        if filename:
            filename_without_extension, extension = filename.split('.')
            self.filename = get_safe_filename(
                f'{filename_without_extension}{filename_suffix}.{extension}'
            )
        else:
            self.filename = get_safe_filename(
                f'{source.title}{filename_suffix}'
                + ('.mp4' if self.output_type.startswith('video') else '.mp3')
            )

        if output_type.startswith('video'):
            self.video = self._get_video_stream(final_definition)
            self.total_size = self.video.filesize
            self.video_bytes_remaining = self.video.filesize

            if output_type == 'videoaudio':
                if self._is_audio_included():
                    self.has_external_audio = False
                else:
                    self._include_external_audio()
                    self.has_external_audio = True

        elif output_type == 'audio':
            self.audio = self._get_audio_stream(final_definition)
            self.total_size = self.audio.filesize
            self.audio_bytes_remaining = self.audio.filesize

        self.formatted_definition = (
            f'[{output_type} '
            if output_type != 'videoaudio'
            else '['
        )

        if definition == 'max':
            self.formatted_definition += (
                f'{self.video.resolution}]'
                if self.output_type.startswith('video')
                else f'{self.audio.abr}]'
            )
        elif is_alias_definition:
            self.formatted_definition += f'{definition}]'
        else:
            self.formatted_definition += f'{final_definition}]'

        self.download_status = DownloadStatusCodes.READY
        self.downloading_stream = None

    def download(self):
        self.download_status = DownloadStatusCodes.DOWNLOADING

        if self.output_type.startswith('video'):
            self.downloading_stream = self.video

            self.video.download(
                output_path=self.output_path,
                filename=f'{temp_filename}[video]',
                skip_existing=False
            )
        if (self.output_type == 'audio'
            or (self.output_type == 'videoaudio' and self.has_external_audio)):
            self.downloading_stream = self.audio

            self.audio.download(
                output_path=self.output_path,
                filename=f'{temp_filename}[audio]',
                skip_existing=False
            )

        self.download_status = DownloadStatusCodes.CONVERTING
        self._convert_dowloaded_resources()
        self.download_status = DownloadStatusCodes.DONE

    def get_total_bytes_remaining(self):
        if self.output_type == 'videoaudio':
            if self.has_external_audio:
                return self.video_bytes_remaining + self.audio_bytes_remaining

            return self.video_bytes_remaining

        if self.output_type == 'videoonly':
            return self.video_bytes_remaining

        if self.output_type == 'audio':
            return self.audio_bytes_remaining

    def _convert_dowloaded_resources(self):
        if self.output_type.startswith('video'):
            if self.output_type == 'videoaudio' and self.has_external_audio:
                self._merge_video_with_external_audio()
            else:
                self._convert_downloaded_video()
        else:
            self._convert_downloaded_audio()

    def _merge_video_with_external_audio(self):
        downloaded_video_extension = self.video.mime_type.split('/')[-1]
        downloaded_audio_extension = self.audio.mime_type.split('/')[-1]

        video_path = path.join(
            self.output_path,
            f'{temp_filename}[video].{downloaded_video_extension}'
        )
        audio_path = path.join(
            self.output_path,
            f'{temp_filename}[audio].{downloaded_audio_extension}'
        )
        output_file_path = path.join(
            self.output_path,
            self.filename
        )

        merge_video_and_audio(video_path, audio_path, output_file_path)

        remove_file(video_path)
        remove_file(audio_path)

    def _convert_downloaded_video(self):
        downloaded_video_extension = self.video.mime_type.split('/')[-1]
        downloaded_temp_filename = (
            f'{temp_filename}[video].{downloaded_video_extension}'
        )

        downloaded_temp_file_path = path.join(
            self.output_path,
            downloaded_temp_filename
        )
        output_file_path = path.join(self.output_path, self.filename)

        options = []
        if self.output_type == 'videoonly':
            options.append(ConversionOptions.NO_AUDIO)

        convert_media(
            downloaded_temp_file_path,
            output_file_path,
            'mp4',
            options=options
        )
        remove_file(downloaded_temp_file_path)

    def _convert_downloaded_audio(self):
        downloaded_audio_extension = self.audio.mime_type.split('/')[-1]
        downloaded_temp_filename = (
            f'{temp_filename}[audio].{downloaded_audio_extension}'
        )

        downloaded_temp_file_path = path.join(
            self.output_path,
            downloaded_temp_filename
        )
        output_file_path = path.join(self.output_path, self.filename)

        convert_media(downloaded_temp_file_path, output_file_path, 'mp3')
        remove_file(downloaded_temp_file_path)

    def _is_audio_included(self):
        return self.video.is_progressive

    def _get_video_stream(self, definition):
        if definition == 'max':
            video_stream = (
                self.source.streams
                    .filter(type='video')
                    .order_by('resolution')[-1]
            )
        else:
            video_stream = (
                self.source.streams
                    .filter(type='video', resolution=definition)[-1]
            )

        return video_stream

    def _get_audio_stream(self, definition):
        if definition == 'max':
            audio_stream = (self.source.streams
                .filter(type='audio')
                .order_by('abr')[-1]
            )
        else:
            audio_stream = (self.source.streams
                .filter(type='audio', abr=definition)[-1]
            )

        return audio_stream

    def _include_external_audio(self):
        self.audio = self._get_audio_stream('max')

        self.total_size += self.audio.filesize
        self.audio_bytes_remaining = self.audio.filesize



from mediakit.streams.screen import screen, ContentCategories
from mediakit.streams.colors import colored, Colors


class MediaKitException(Exception):
    def show_message(self):
        screen.append_content(self.message, self.category)


class CommandUnavailable(MediaKitException):
    def __init__(self, command):
        self.message = (
            f"Command '{command}' is not available.\n\n"
        )
        self.category = ContentCategories.ERROR

        super().__init__(self.message)


class FFMPEGNotAvailable(MediaKitException):
    def __init__(self):
        self.message = (
            'FFmpeg is a required package for MediaKit, '
            "but it doesn't appear to be installed.\n"
            'You can install FFmpeg at https://ffmpeg.org/download.html.\n\n'
        )
        self.category = ContentCategories.WARNING

        super().__init__(self.message)


class InvalidVideoURLError(MediaKitException):
    def __init__(self):
        self.message = (
            'Could not recognize the provided URL. Please, try again.\n\n'
        )
        self.category = ContentCategories.ERROR

        super().__init__(self.message)


class UnspecifiedError(MediaKitException):
    def __init__(self):
        self.message = 'Something went wrong. :(\nPlease, try again.\n\n'
        self.category = ContentCategories.ERROR

        super().__init__(self.message)


class NoAvailableSpecifiedFormats(MediaKitException):
    def __init__(self, available_formats):
        self.message = (
            'None of the specified formats were found.\n'
            + 'Please, verify your entries and try again.\n\n'

            + 'The formats available for this video are:\n'
            + colored(
                '   (video) ',
                style=Colors.style.DIM + Colors.style.BRIGHT
            )
            + colored(
                ' '.join(available_formats['video']),
                fore=Colors.fore.BLUE,
                style=Colors.style.BRIGHT
            )
            + colored(
                '\n   (audio) ',
                style=Colors.style.DIM + Colors.style.BRIGHT
            )
            + colored(
                ' '.join(available_formats['audio']),
                fore=Colors.fore.BLUE,
                style=Colors.style.BRIGHT
            )
            + '\n\n'
        )
        self.category = ContentCategories.ERROR

        super().__init__(self.message)



from math import ceil

from clint.textui.cols import console_width

from mediakit.utils.format import len_ansi_safe
from .colors import colored, Colors


class ANSIConsoleExpressions:
    MOVE_CURSOR_ONE_LINE_UP = '\033[A'
    MOVE_CURSOR_TO_BEGGINING_OF_LINE = '\033[0G'
    CLEAR_CHARACTERS_TO_THE_RIGHT_OF_CURSOR = '\033[K'
    CLEAR_LINE = (
        MOVE_CURSOR_TO_BEGGINING_OF_LINE
        + CLEAR_CHARACTERS_TO_THE_RIGHT_OF_CURSOR
    )


class ContentCategories:
    NORMAL = 'NORMAL'
    INFO = 'INFO'
    WARNING = 'WARNING'
    ERROR = 'ERROR'
    USER_INPUT = 'USER_INPUT'


class Content:
    def __init__(self, index_on_screen, text, category):
        self.index_on_screen = index_on_screen

        if category == ContentCategories.NORMAL:
            self.category_label = ''
        elif category == ContentCategories.INFO:
            self.category_label = colored('info ', fore=Colors.fore.BLUE)
        elif category == ContentCategories.WARNING:
            self.category_label = colored('warning ', fore=Colors.fore.YELLOW)
        elif category == ContentCategories.ERROR:
            self.category_label = colored('error ', fore=Colors.fore.RED)
        elif category == ContentCategories.USER_INPUT:
            self.category_label = colored('? ', fore=Colors.fore.BLUE)

        self.update_inner_text(text)

    def update_inner_text(self, new_text):
        new_text_stripped_of_leading_newlines = new_text.lstrip('\n')
        number_of_leading_newlines = (
            len(new_text) - len(new_text_stripped_of_leading_newlines)
        )

        self.inner_text = (
            '\n' * number_of_leading_newlines
            + self.category_label
            + new_text_stripped_of_leading_newlines
        )

class Screen:
    def __init__(self):
        self.contents = []
        self.prompt_message = None

    def append_content(self, content_text, category=ContentCategories.NORMAL):
        index_on_screen = len(self.contents)

        new_content = Content(index_on_screen, content_text, category)

        self.contents.append(new_content)
        self._render_content(new_content)

        return new_content

    def update_content(self, content, new_content_text):
        self._clear_lines_starting_at(content)

        content.update_inner_text(new_content_text)

        self._render_contents_starting_at(content)

    def remove_content(self, content):
        index_to_remove_at = content.index_on_screen

        self._clear_lines_starting_at(content)
        self.contents.pop(index_to_remove_at)

        base_index = index_to_remove_at

        for index in range(base_index, len(self.contents)):
            self.contents[index].index_on_screen = index

        needs_to_rerender_contents = (
            len(self.contents) > 0
            and base_index < len(self.contents)
        )

        if needs_to_rerender_contents:
            self._render_contents_starting_at(self.contents[base_index])

    def clear_lines(self, number_of_lines_to_clear):
        clear_expression = (
            ANSIConsoleExpressions.MOVE_CURSOR_ONE_LINE_UP.join(
                [ANSIConsoleExpressions.CLEAR_LINE] * number_of_lines_to_clear
            )
        )

        print(clear_expression, end='')

    def get_console_width(self):
        return console_width({})

    def prompt(
        self,
        message,
        valid_inputs=[],
        invalid_inputs=[],
        case_sensitive=False
    ):
        if not case_sensitive:
            for i in range(len(valid_inputs)):
                valid_inputs[i] = valid_inputs[i].lower()

        valid_inputs = set(valid_inputs)
        invalid_inputs = set(invalid_inputs)

        self.prompt_message = self.append_content(
            message,
            category=ContentCategories.USER_INPUT
        )

        while True:
            entry = input()

            if not case_sensitive:
                entry = entry.lower()

            valid_entry = (
                (entry in valid_inputs if len(valid_inputs) > 0 else True)
                and entry not in invalid_inputs
            )

            if valid_entry:
                return entry

            self._erase_prompt_entry()

    def _render_content(self, content):
        print(content.inner_text, end='')

    def _clear_lines_starting_at(self, content):
        current_console_width = self.get_console_width()

        lines_to_clear = 0
        for i in range(content.index_on_screen, len(self.contents)):
            content = self.contents[i]

            lines_to_clear += self._count_lines_occupied_by(
                content,
                current_console_width
            )

        self.clear_lines(lines_to_clear)

    def _render_contents_starting_at(self, content):
        for i in range(content.index_on_screen, len(self.contents)):
            self._render_content(self.contents[i])

    def _erase_prompt_entry(self):
        lines_occupied_by_entry = 2 # <entry>\n<empty line> -> 2 lines to clear
        self.clear_lines(lines_occupied_by_entry)

        self._clear_lines_starting_at(self.prompt_message)
        self._render_contents_starting_at(self.prompt_message)

    def _count_lines_occupied_by(self, content, current_console_width):
        lines = content.inner_text.split('\n')

        lines_occupied = 0
        for i in range(len(lines)):
            line = lines[i]

            lines_occupied += max(
                ceil(len_ansi_safe(line) / current_console_width),
                1
            )

        return lines_occupied

screen = Screen()



from threading import Thread
from time import sleep
from math import floor

from mediakit.info import name, version
from mediakit.streams.screen import screen, ContentCategories
from mediakit.streams.colors import colored, Colors
from mediakit.media.download import MediaResource, DownloadStatusCodes
from mediakit.utils.format import limit_text_length, parse_int
from mediakit.constants import (
    DOWNLOAD_FORMATS,
    VIDEO_DEFINITIONS,
    VIDEO_DEFINITIONS_ALIASES,
    AUDIO_DEFINITIONS,
    AVAILABLE_DEFINITIONS
)
from mediakit import exceptions

loading_dots = ['⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏']
rotating_lines = ['|', '/', '-', '\\']


class DownloadCLI:
    def __init__(self):
        self.max_width = 80

        self.loading_label = None

        self.video = None
        self.output_path = None
        self.filename = None
        self.short_video_title = None

        self.available_formats = { 'video': [], 'audio': [] }
        self.media_resources_to_download = []
        self.skipped_formats = []
        self.formats_replaced_by_fallback = []

        self.downloading_media_resource = None
        self.download_progress_ui = None
        self.rotating_line_frame = 0
        self.loading_dots_frame = 0

        self.ui_update_intervals = {
            DownloadStatusCodes.DOWNLOADING: 0.2,
            DownloadStatusCodes.CONVERTING: 0.1
        }
        self.default_ui_update_interval = 0.2
        self.progress_ui_update_interval = self.default_ui_update_interval

        self.is_terminated = False

    def start(self):
        self._show_header()

    def register_download_info(self, video, output_path, filename, formats):
        self.video = video
        self.output_path = output_path
        self.filename = filename

        self.available_formats = self._get_available_formats()
        self.media_resources_to_download = (
            self._get_media_resources_to_download(formats)
        )

        self.short_video_title = self._format_video_title(26)

        self.remove_loading_label()

    def show_video_heading(self):
        formatted_video_length = self._format_video_length()

        remaining_number_of_characters = (
            min(screen.get_console_width(), self.max_width)
            - len(formatted_video_length)
            - 4
        )

        formatted_video_title = (
            self._format_video_title(remaining_number_of_characters)
        )

        spaces_between = ' ' * (
            min(screen.get_console_width(), self.max_width)
            - len(formatted_video_title)
            - len(formatted_video_length)
            - 3
        )

        video_summary = (
            colored(
                formatted_video_title,
                fore=Colors.fore.CYAN,
                style=Colors.style.BRIGHT
            )
            + spaces_between
            + colored(
                f'({formatted_video_length})\n\n',
                style=Colors.style.BRIGHT
            )
        )

        screen.append_content(video_summary)

    def show_download_summary(self):
        if self.were_all_formats_skipped():
            raise exceptions.NoAvailableSpecifiedFormats(self.available_formats)

        if len(self.skipped_formats) > 0:
            self._show_skipped_formats_warning()

        if len(self.formats_replaced_by_fallback) > 0:
            self._show_fallback_replacement_summary()

        formatted_title = limit_text_length(self.video.title, 26)
        formatted_download_formats = self._get_formatted_download_formats()

        media_resource_sizes = map(
            lambda media_resource: media_resource.total_size,
            self.media_resources_to_download
        )
        total_download_size = sum(media_resource_sizes)
        formatted_download_size = self._format_data_size(total_download_size)

        is_preceded_by_format_warnings = (
            len(self.skipped_formats) > 0
            or len(self.formats_replaced_by_fallback) > 0
        )

        screen.append_content(
            ('\n' if is_preceded_by_format_warnings else '')
            + 'Ready to download '
            + colored(
                f'{formatted_title} ',
                fore=Colors.fore.CYAN,
                style=Colors.style.BRIGHT
            )
            + colored(
                f'{formatted_download_formats}\n',
                fore=Colors.fore.BLUE,
                style=Colors.style.BRIGHT
            ),
            ContentCategories.INFO
        )
        screen.append_content(
            'Total download size: '
            + colored(
                f'{formatted_download_size}\n\n',
                fore=Colors.fore.YELLOW
            ),
            ContentCategories.INFO
        )

    def ask_for_confirmation_to_download(self):
        prompt_message = (
            'Confirm download? '
            + colored(
                '(Y/n) ',
                fore=Colors.fore.CYAN,
            )
        )

        answer = screen.prompt(
            prompt_message,
            valid_inputs=['', 'y', 'n']
        )

        return answer != 'n'

    def download_selected_formats(self):
        for media_resource in self.media_resources_to_download:
            self._create_download_progress(media_resource)
            screen_thread = Thread(
                target=self._keep_download_progress_ui_updated
            )
            screen_thread.start()

            media_resource.download()

            self.end_download_progress()

    def update_download_progress_info(self, bytes_remaining):
        downloading_stream = self.downloading_media_resource.downloading_stream

        if self.downloading_media_resource.output_type == 'videoaudio':
            if downloading_stream is self.downloading_media_resource.video:
                self.downloading_media_resource.video_bytes_remaining = (
                    bytes_remaining
                )
            elif downloading_stream is self.downloading_media_resource.audio:
                self.downloading_media_resource.audio_bytes_remaining = (
                    bytes_remaining
                )

        if self.downloading_media_resource.output_type == 'videoonly':
            self.downloading_media_resource.video_bytes_remaining = (
                bytes_remaining
            )

        if self.downloading_media_resource.output_type == 'audio':
            self.downloading_media_resource.audio_bytes_remaining = (
                bytes_remaining
            )

    def end_download_progress(self):
        self._update_download_progress_ui()

        self.downloading_media_resource = None
        self.download_progress_ui = None

    def show_success_message(self):
        screen.append_content(
            colored(
                '\nSuccess! ',
                fore=Colors.fore.GREEN
            )
            + 'Files saved at '
            + colored(
                f'{self.output_path}/\n\n',
                fore=Colors.fore.CYAN,
            )
        )

    def terminate(self):
        self.is_terminated = True

    def remove_loading_label(self):
        if self.loading_label is not None:
            screen.remove_content(self.loading_label)

    def were_all_formats_skipped(self):
        return (
            len(self.media_resources_to_download) == 0
            and len(self.skipped_formats) > 0
        )

    def _show_header(self):
        screen.append_content(colored(
            f'{name.lower()} v{version}\n\n',
            style=Colors.style.BRIGHT
        ))

        self.loading_label = screen.append_content(
            'Loading video.\n\n',
            ContentCategories.INFO
        )

        number_of_dots = 1

        def run_loading_animation():
            nonlocal number_of_dots

            while self.video is None and not self.is_terminated:
                number_of_dots = (number_of_dots + 1) % 4

                if self.loading_label is None:
                    break

                screen.update_content(
                    self.loading_label,
                    'Loading video' + '.' * number_of_dots + '\n\n'
                )

                sleep(0.4)

        Thread(target=run_loading_animation).start()

    def _get_media_resources_to_download(self, formats):
        if len(formats) == 0:
            default_media_resource = MediaResource(
                self.video,
                'videoaudio',
                output_path=self.output_path,
                filename=self.filename
            )

            return [default_media_resource]

        media_resources_to_download = []

        lowecased_formats = [
            selected_format.lower()
            for selected_format in formats
        ]
        grouped_formats = self._group_and_validate_formats(lowecased_formats)

        definitions_count = {}
        for group in grouped_formats:
            download_format = group['format']
            current_count = definitions_count.get(download_format, 0)
            definitions_count[download_format] = current_count + 1

        for group in grouped_formats:
            download_format = group['format']
            definition = group['definition']

            should_append_format_to_filename = (
                definitions_count[download_format] > 1
            )

            available_definition = self._get_available_definition(
                download_format,
                definition
            )

            if available_definition is None:
                self.skipped_formats.append([download_format, definition])
                continue

            if available_definition != definition:
                self.formats_replaced_by_fallback.append({
                    'base': {
                        'format': download_format,
                        'definition': definition
                    },
                    'fallback': {
                        'format': download_format,
                        'definition': available_definition
                    },
                })

            if should_append_format_to_filename:
                filename_suffix = (
                    (
                        f' [{download_format}] '
                        if download_format != 'videoaudio'
                        else ''
                    )
                    + (
                        f' [{available_definition}]'
                        if available_definition != 'max'
                        else ''
                    )
                )
            else:
                filename_suffix = ''

            media_resource = MediaResource(
                self.video,
                download_format,
                output_path=self.output_path,
                definition=available_definition,
                filename=self.filename,
                filename_suffix=filename_suffix
            )

            media_resources_to_download.append(media_resource)

        return media_resources_to_download

    def _group_and_validate_formats(self, formats):
        grouped_formats = []

        i = 0
        while i < len(formats):
            current_item = formats[i]
            next_item = formats[i + 1] if i < len(formats) - 1 else None

            is_current_item_a_format = current_item in DOWNLOAD_FORMATS
            is_current_item_a_definition = current_item in AVAILABLE_DEFINITIONS
            is_next_item_a_format = next_item in DOWNLOAD_FORMATS
            is_next_item_a_definition = next_item in AVAILABLE_DEFINITIONS

            if is_current_item_a_format:
                if is_next_item_a_definition:
                    grouped_formats.append({
                        'format': current_item,
                        'definition': next_item
                    })
                    i += 1
                elif next_item is None or is_next_item_a_format:
                    grouped_formats.append({
                        'format': current_item,
                        'definition': 'max'
                    })
                else:
                    self.skipped_formats.append([current_item, next_item])
                    i += 1

            elif is_current_item_a_definition:
                grouped_formats.append({
                    'format': 'videoaudio',
                    'definition': current_item
                })

            else:
                self.skipped_formats.append([current_item])

            i += 1

        return grouped_formats

    def _get_available_definition(self, base_format, base_definition):
        if base_format.startswith('video'):
            is_valid_definition = (
                base_definition in VIDEO_DEFINITIONS
                or base_definition in VIDEO_DEFINITIONS_ALIASES
            )
        else:
            is_valid_definition = base_definition in AUDIO_DEFINITIONS

        if not is_valid_definition:
            return None

        possible_definition = base_definition
        while possible_definition is not None:
            if self._is_definition_available(base_format, possible_definition):
                return possible_definition

            possible_definition = (
                VIDEO_DEFINITIONS_ALIASES
                    .get(possible_definition, possible_definition)
            )

            possible_definition = (
                VIDEO_DEFINITIONS[possible_definition]['next']
                if base_format.startswith('video')
                else AUDIO_DEFINITIONS[possible_definition]['next']
            )

        return None

    def _is_definition_available(self, base_format, definition):
        if base_format.startswith('video'):
            final_definition = (
                VIDEO_DEFINITIONS_ALIASES.get(definition, definition)
            )

            if final_definition == 'max':
                video_streams = self.video.streams.filter(type='video')

                return len(video_streams) > 0

            video_streams_with_specified_resolution = (
                self.video.streams.filter(
                    type='video',
                    resolution=final_definition
                )
            )

            return len(video_streams_with_specified_resolution) > 0

        else:
            if definition == 'max':
                audio_streams = self.video.streams.filter(type='audio')

                return len(audio_streams) > 0

            audio_streams_with_specified_resolution = (
                self.video.streams.filter(
                    type='audio',
                    abr=definition
                )
            )

            return len(audio_streams_with_specified_resolution) > 0

    def _create_download_progress(self, media_resource):
        self.downloading_media_resource = media_resource

        self.download_progress_ui = screen.append_content(
            self._get_current_download_progress()
        )

    def _update_download_progress_ui(self):
        screen.update_content(
            self.download_progress_ui,
            self._get_current_download_progress()
        )

    def _keep_download_progress_ui_updated(self):
        while self.downloading_media_resource and not self.is_terminated:
            self._update_download_progress_ui()
            self._update_progress_ui_interval_if_necessary()

            sleep(self.progress_ui_update_interval)

    def _update_progress_ui_interval_if_necessary(self):
        download_status = self.downloading_media_resource.download_status
        required_interval = self.ui_update_intervals.get(
            download_status,
            self.default_ui_update_interval
        )

        if self.progress_ui_update_interval != required_interval:
            self.progress_ui_update_interval = required_interval

    def _get_current_download_progress(self):
        download_status = self.downloading_media_resource.download_status

        status_character, status_color = (
            self._get_status_resources(download_status)
        )

        heading = self._get_download_progress_heading(
            download_status,
            status_character,
            status_color
        )

        current_download_progress = heading + '\n'

        if download_status == DownloadStatusCodes.DONE:
            return current_download_progress

        progress_bar = self._get_download_progress_bar(status_color)

        current_download_progress += '\n' + progress_bar + '\n\n'

        return current_download_progress

    def _get_download_progress_heading(
        self,
        download_status,
        status_character,
        status_color
    ):
        label = download_status.title()

        is_downloading_first_media_resource = (
            self.downloading_media_resource
            is self.media_resources_to_download[0]
        )

        should_include_starting_newline = (
            is_downloading_first_media_resource
            or (download_status != DownloadStatusCodes.DONE)
        )

        heading = (
            ('\n' if should_include_starting_newline else '')
            + colored(
                f'{status_character} {label} ',
                fore=status_color
            )
            + colored(
                f"{self.short_video_title} ",
                style=Colors.style.BRIGHT
            )
            + colored(
                f'{self.downloading_media_resource.formatted_definition}',
                fore=Colors.fore.BLUE,
                style=Colors.style.BRIGHT
            )
        )

        return heading

    def _get_download_progress_bar(self, status_color):
        media_resource = self.downloading_media_resource

        bytes_downloaded = (
            media_resource.total_size
            - media_resource.get_total_bytes_remaining()
        )

        progress_percentage = bytes_downloaded / media_resource.total_size
        available_width = min(screen.get_console_width(), self.max_width)

        formatted_bytes_downloaded = self._format_data_size(bytes_downloaded)
        formatted_total_size = self._format_data_size(media_resource.total_size)

        progress_bar_right = (
            colored(
                f' {progress_percentage * 100:.1f}% ',
                fore=status_color,
                style=Colors.style.BRIGHT
            )
            + colored(
                f'({formatted_bytes_downloaded} / '
                + f'{formatted_total_size})',
                style=Colors.style.DIM
            )
        )

        max_progress_bar_right_width = len(
            f' 100.0% ({formatted_total_size} / {formatted_total_size})'
        )

        available_space_for_loading_bar = (
            available_width - max_progress_bar_right_width - 2
        )
        loaded_section_length = floor(
            available_space_for_loading_bar * progress_percentage
        )

        progress_bar_left = (
            '  '
            + colored(
                '█' * loaded_section_length,
                fore=status_color
            )
            + colored(
                '█' * (available_space_for_loading_bar - loaded_section_length),
                style=Colors.style.DIM
            )
        )

        progress_bar = progress_bar_left + progress_bar_right

        return progress_bar

    def _format_video_length(self):
        seconds_section = self.video.length % 60
        remaining_minutes = self.video.length // 60
        minutes_section = remaining_minutes % 60
        hours_section = remaining_minutes // 60

        formatted_video_length = (
            (f'{hours_section}:' if hours_section > 0 else '')
            + f'{minutes_section:0>2}:'
            + f'{seconds_section:0>2}'
        )

        return formatted_video_length

    def _format_video_title(self, limit_of_characters):
        return limit_text_length(self.video.title, limit_of_characters)

    def _format_data_size(self, size_in_bytes):
        size_in_megabytes = size_in_bytes / 1000000

        return f'{size_in_megabytes:.1f} MB'

    def _get_formatted_download_formats(self):
        formatted_definitions = map(
            lambda media_resource: media_resource.formatted_definition,
            self.media_resources_to_download
        )

        formatted_download_formats = ' '.join(formatted_definitions)

        return formatted_download_formats

    def _get_status_resources(self, download_status):
        status_character, status_color = '', ''

        if download_status == DownloadStatusCodes.DOWNLOADING:
            self.rotating_line_frame = (
                (self.rotating_line_frame + 1) % 4
            )

            status_character = rotating_lines[self.rotating_line_frame]
            status_color = Colors.fore.YELLOW

        if download_status == DownloadStatusCodes.CONVERTING:
            self.loading_dots_frame = (
                (self.loading_dots_frame + 1) % 10
            )

            status_character = loading_dots[self.loading_dots_frame]
            status_color = Colors.fore.CYAN

        if download_status == DownloadStatusCodes.DONE:
            status_character = '✔'
            status_color = Colors.fore.GREEN

        return status_character, status_color

    def _get_available_formats(self):
        video_streams = self.video.streams.filter(type='video')
        audio_streams = self.video.streams.filter(type='audio')

        available_formats = {
            'video': set(),
            'audio': set(),
        }
        for stream in video_streams:
            if stream.resolution is not None:
                available_formats['video'].add(stream.resolution)
        for stream in audio_streams:
            available_formats['audio'].add(stream.abr)


        available_formats_sorted_by_definition = {
            'video': reversed(
                sorted(available_formats['video'], key=parse_int)
            ),
            'audio': reversed(
                sorted(available_formats['audio'], key=parse_int)
            )
        }

        return available_formats_sorted_by_definition

    def _show_skipped_formats_warning(self):
        formatted_skipped_formats = [
            f'[{" ".join(skipped_format)}]'
            for skipped_format in self.skipped_formats
        ]

        skipped_formats_message = (
            ('Formats ' if len(self.skipped_formats) > 1 else 'Format ')
            + colored(
                ' '.join(formatted_skipped_formats),
                fore=Colors.fore.MAGENTA,
                style=Colors.style.BRIGHT
            )
            + (' were ' if len(self.skipped_formats) > 1 else ' was ')
            + 'not found. Skipping '
            + ('them' if len(self.skipped_formats) > 1 else 'it')
            + '...\n'
        )

        screen.append_content(
            skipped_formats_message,
            ContentCategories.WARNING
        )

    def _show_fallback_replacement_summary(self):
        def get_formatted_replaced_format(current_format, definition):
            return (
                '['
                + (
                    f'{current_format} '
                    if current_format != 'videoaudio'
                    else ''
                )
                + f'{definition}]'
            )

        replacement_summaries = []
        for replaced_format in self.formats_replaced_by_fallback:
            base_format = replaced_format['base']['format']
            base_definition = replaced_format['base']['definition']
            fallback_format = replaced_format['fallback']['format']
            fallback_definition = replaced_format['fallback']['definition']

            formatted_base = get_formatted_replaced_format(
                base_format,
                base_definition
            )
            formatted_fallback = get_formatted_replaced_format(
                fallback_format,
                fallback_definition
            )

            replacement_summaries.append(
                'Format '
                + colored(
                    formatted_base,
                    fore=Colors.fore.YELLOW,
                    style=Colors.style.BRIGHT
                )
                + ' is not available for this video. Falling back to '
                + colored(
                    formatted_fallback,
                    fore=Colors.fore.BLUE,
                    style=Colors.style.BRIGHT
                )
                + '\n'
            )

        for summary in replacement_summaries:
            screen.append_content(summary, ContentCategories.WARNING)



import re

def limit_text_length(text, limit_of_caracters):
    if len(text) > limit_of_caracters:
        partial_text_with_ellipsis = f'{text[:limit_of_caracters - 3]}...'

        return partial_text_with_ellipsis

    return text

def len_ansi_safe(string):
    ansi_escape_codes_regex = r'\x1b\[[;\d]*[A-Za-z]'

    string_striped_of_ansi_codes = re.sub(
        ansi_escape_codes_regex,
        '',
        string
    )

    return len(string_striped_of_ansi_codes)

def parse_int(string):
    integer_section_as_list = []
    for character in string:
        if character.isnumeric():
            integer_section_as_list.append(character)
        else:
            break

    integer_section = ''.join(integer_section_as_list)

    return int(integer_section)



from os import path
import re

from mediakit.utils.commands import run_command_in_background


def file_exists(file_path):
    return path.isfile(file_path)


def move_file(file_path, new_file_path):
    move_command = f'mv "{file_path}" "{new_file_path}"'

    run_command_in_background(move_command)


def remove_file(file_path):
    remove_command = f'rm -rf "{file_path}"'

    run_command_in_background(remove_command)


def get_safe_filename(filename):
    partially_safe_filename = re.sub(r'[{}\\"\']', '', filename)
    safe_filename = re.sub(r'[/]', ' ', partially_safe_filename)

    return safe_filename


def get_filename_from(file_path):
    possible_filename = path.basename(file_path)
    valid_filename = (
        0 < possible_filename.rfind('.') < len(possible_filename) - 1
    )

    return possible_filename if valid_filename else ''


def increment_filename_if_exists(file_path):
    file_path_dirname = path.dirname(file_path)

    filename = get_filename_from(file_path)
    extension_start_index = filename.rfind('.')

    filename_without_extension = filename[:extension_start_index]
    extension = filename[extension_start_index+1:]

    final_file_path = file_path
    filename_number = 0
    while True:
        if not file_exists(final_file_path):
            return final_file_path

        filename_number += 1
        final_file_path = path.join(
            file_path_dirname,
            f'{filename_without_extension}({filename_number}).{extension}'
        )



from os import path

from setuptools import setup, find_packages

from mediakit import info

current_dirname = path.dirname(path.abspath(__file__))
readme_path = path.join(current_dirname, 'README.md')

with open(readme_path, encoding='utf-8') as readme:
    long_description = readme.read()

setup(
    name=info.name.lower(),
    version=info.version,
    description=info.description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=info.author,
    author_email=info.author_email,
    url=info.url,
    download_url=info.download_url,
    license=info.license,
    packages=find_packages(),
    include_package_data=True,
    keywords=[
        'youtube',
        'media',
        'video',
        'audio',
        'download',
        'convert',
        'command line',
        'cli'
    ],
    python_requires='>=3.6',
    install_requires=[
        'pytube>=10.0.0',
        'clint',
        'imageio',
        'imageio-ffmpeg',
        'colorama'
    ],
    entry_points={
        'console_scripts': [
            'mediakit=mediakit.actions:download'
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: Unix',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',

        'Topic :: Multimedia :: Video',
        'Topic :: Multimedia :: Video :: Conversion',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Sound/Audio :: Conversion'
    ]
)



from os import path

from pytube import YouTube
from pytube.exceptions import RegexMatchError as PytubeRegexMatchError

from mediakit.streams.arguments import parse_download_arguments
from mediakit.streams.cli import DownloadCLI
from mediakit.utils.files import get_filename_from
from mediakit.constants import FFMPEG_BINARY
from mediakit import exceptions


def download():
    def on_download_progress(stream, chunk, bytes_remaining):
        download_cli.update_download_progress_info(bytes_remaining)

    arguments = parse_download_arguments()
    video_url = arguments.video_url
    output_path = path.abspath(arguments.output_path)

    filename = get_filename_from(output_path)
    if filename:
        output_path = path.dirname(output_path)

    formats = arguments.formats

    download_cli = DownloadCLI()
    download_cli.start()

    try:
        if not FFMPEG_BINARY:
            raise exceptions.FFMPEGNotAvailable()

        video = YouTube(video_url)
        download_cli.register_download_info(
            video,
            output_path,
            filename,
            formats
        )
        download_cli.show_video_heading()
        download_cli.show_download_summary()

        confirmed = download_cli.ask_for_confirmation_to_download()
        if not confirmed:
            return

        video.register_on_progress_callback(on_download_progress)

        download_cli.download_selected_formats()
        download_cli.show_success_message()

    except exceptions.FFMPEGNotAvailable as exception:
        download_cli.remove_loading_label()
        exception.show_message()
    except exceptions.NoAvailableSpecifiedFormats as exception:
        exception.show_message()
    except PytubeRegexMatchError:
        download_cli.remove_loading_label()
        exceptions.InvalidVideoURLError().show_message()
    except KeyboardInterrupt:
        pass
    except Exception:
        download_cli.remove_loading_label()
        exceptions.UnspecifiedError().show_message()
    finally:
        download_cli.terminate()
